"""Base backends for letters."""
import smtplib
from dataclasses import dataclass
from email.message import EmailMessage
from typing import Any, Dict, Optional, cast
from uuid import UUID

from filed import File

from messenger.constants import TransportMethod
from messenger.models import Address, Letter

from .base import BaseBackend, FileBackendMixin, MemoryBackendMixin
from .email import SmtpBackendMixin


@dataclass
class LetterMessage:
    """A letter message prepared for send.

    Attributes:
        uuid: The message UUID.
        recipient: The address of the message recipient.
        file: The file with the actual letter.
        sender: The sender address.
    """

    uuid: UUID
    recipient: Address
    file: File
    sender: Optional[Address] = None


class BaseLetterBackend(FileBackendMixin, BaseBackend):
    """Base backend for all letter backends.

    Attributes:
        sender: The sender address.
        user: The user for authentication at the provider.
        password: The password for authentication at the provider.
        test: Whether a test mode should be used.

    Behaviour of `user`, `password` and `test` arguments is implementation specific.
    """

    method = TransportMethod.LETTER

    def __init__(self, *, sender: Optional[Dict[str, str]] = None, user: Optional[str] = None,
                 password: Optional[str] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if sender is None:
            self.sender = None
        else:
            try:
                self.sender = Address(**sender)
            except TypeError as error:
                raise ValueError("Invalid sender: {}".format(error)) from error
        self.user = user
        self.password = password

    def get_message(self, model: Letter) -> LetterMessage:
        """Return formatted letter message from Letter object."""
        return LetterMessage(uuid=model.uuid, recipient=model.recipient, file=self.get_file(model.file),
                             sender=self.sender)


class MemoryLetterBackend(MemoryBackendMixin, BaseLetterBackend):
    """Backend which stores sent letters in memory.

    This backend is not intended for production use, but only for testing and development.
    """


class EmailLetterBackend(SmtpBackendMixin, BaseLetterBackend):
    """A letter backend which sends letters as emails.

    Attributes:
        to_address: The recipient(s) email address.
        from_address: The sender email address.
        subject: An email subject.
        body: An email body.

    Subject and body can use `{}`-format syntax and `letter` variable with a ``LetterMessage``.
    """

    def __init__(self, *, to_address: str, from_address: str = 'webmaster@localhost',
                 subject: str = 'A letter from messenger',
                 body: str = 'A letter to {letter.recipient}, see attachment.', **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.to_address = to_address
        self.from_address = from_address
        self.subject = subject
        self.body = body

    def _send(self, message: LetterMessage) -> None:
        """Actually send the letter.

        Raises:
            RuntimeError: An error occures while sending the letter.
        """
        email = EmailMessage()
        email['From'] = self.from_address
        email['To'] = self.to_address
        email['Subject'] = self.subject.format(letter=message)
        email.set_content(self.body.format(letter=message))
        email.add_attachment(message.file.read(), *cast(str, message.file.mimetype).split('/', 1),
                             filename=message.file.name)
        try:
            self.open()
            cast(smtplib.SMTP, self.connection).send_message(email)
        except smtplib.SMTPException as error:
            raise RuntimeError('Sending failed with error: {}'.format(error))
