"""Command utilities."""
from datetime import datetime
from enum import IntEnum, unique
from functools import wraps
from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union, cast

import click
from click import echo as _echo
from pytz import UTC
from typing_extensions import Literal

import messenger
from messenger.constants import TransportMethod
from messenger.settings import SETTINGS


@unique
class Verbosity(IntEnum):
    """Verbosity levels."""

    SILENT = 0
    INFO = 1
    DETAIL = 2
    DEBUG = 3


def echo(message: str, verbosity: Verbosity, *, nl: bool = True, err: bool = False,
         log: Optional[Callable[[str], None]] = None) -> None:
    """Print message with respect to verbosity level.

    Arguments:
        message: The message to print.
        verbosity: Verbosity of the message.
        nl: Whether to print newline after message.
        err: Whether to print the message to error output instead.
        log: Callable to log the message as well.
    """
    if verbosity <= SETTINGS['_verbosity']:
        _echo(message, nl=nl, err=err)
    if log is not None:
        log(message)


F = TypeVar('F', bound=Callable[..., Any])


def messenger_command(command: Callable = click.command) -> Callable:
    """Turn function into a messenger command.

    Arguments:
        command: A click `command` callable to attach command to a group.
    """
    def decorator(func: F) -> F:
        # This decorator turns the function into a click command. To keep a reasonable order of options, that is
        # end usage with --connection, --config, --verbosity, --version and --help in this order,
        # we need to keep this order of click decorators.
        @click.version_option(version=messenger.__version__, message="fred-messenger %(version)s")
        @click.option('-v', '--verbosity', type=int, default=1, show_default=True, help="Set verbosity level.")
        @click.option('--config', type=click.Path(exists=True, dir_okay=False), help="Set custom config file.")
        @click.option('--connection', help="Set custom database connection string.")
        @command()
        @wraps(func)
        def wrapped(*, config: str, connection: str, verbosity: int, **kwargs: Any) -> Any:
            SETTINGS.load(config)
            SETTINGS['_verbosity'] = verbosity
            if connection:
                SETTINGS['db_connection'] = connection
            return func(**kwargs)

        return cast(F, wrapped)
    return decorator


def _parse_methods(ctx: click.Context, param: click.Parameter, value: Sequence[Union[Literal['all'], TransportMethod]]
                   ) -> Tuple[TransportMethod, ...]:
    """Return selected collection of transport methods."""
    if 'all' in value:
        return tuple(TransportMethod)
    else:
        # Only TransportMethod values and 'all' are allowed choices.
        return tuple(TransportMethod(v) for v in value)


def method_option(func: Callable) -> Callable:
    """Add --type option to a command."""
    # Use `methods` as the name of the variable.
    option = click.option(
        '--type', 'methods', type=click.Choice(('all', ) + tuple(str(m) for m in TransportMethod)), multiple=True,
        default=('all', ), show_default=True, callback=_parse_methods,
        help="Run only service with selected type. May be set multiple times.")
    return option(func)


def _make_aware_callback(ctx: click.Context, param: click.Parameter, value: Optional[datetime]) -> Optional[datetime]:
    """Make the datetime aware."""
    if value is None:
        return value
    return UTC.localize(value)


def limits_options(func: Callable) -> Callable:
    """Add --from and --to options to a command."""
    from_option = click.option(
        '--from', 'from_datetime', type=click.DateTime(), callback=_make_aware_callback,
        help="Set minimal date and time limit for messages in UTC timezone.")
    to_option = click.option(
        '--to', 'to_datetime', type=click.DateTime(), callback=_make_aware_callback,
        help="Set maximal date and time limit for messages in UTC timezone.")
    return from_option(to_option(func))
