"""Database utilities."""
from contextlib import contextmanager
from functools import lru_cache
from typing import Type, cast

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, sessionmaker

from .settings import SETTINGS


@lru_cache()
def get_engine() -> Engine:
    """Return engine as defined by settings."""
    return create_engine(SETTINGS['db_connection'], echo=SETTINGS['db_echo'])


@lru_cache()
def get_session_cls() -> Type[Session]:
    """Return session class for defined engine."""
    return cast(Type[Session], sessionmaker(bind=get_engine()))


@contextmanager
def atomic() -> Session:
    """Context manager which wraps content into a transaction."""
    session = get_session_cls()()
    try:
        yield session
        session.commit()
    except BaseException:
        session.rollback()
        raise
    finally:
        session.close()
