"""Message filters."""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Container, Optional, Union

from .models import Record
from .utils import load_instance


class Filter(ABC):
    """Abstract base class for filters."""

    @abstractmethod
    def filter(self, model: Record) -> bool:
        """Return whether message instance matches filter."""


class NegativeFilter(Filter):
    """Filter which negates other filters.

    Attributes:
        inner_filter: The negated filter.
    """

    def __init__(self, inner_filter: Union[str, Filter], **kwargs: Any):
        """Wrap the provided filter, load if required.

        Arguments:
            inner_filter: An instance or dotted-path to a filter.
            **kwargs: Other arguments to initialize inner filter from dotted-path.
        """
        if isinstance(inner_filter, str):
            self.inner_filter = load_instance(inner_filter, **kwargs)
        else:
            self.inner_filter = inner_filter

    def filter(self, model: Record) -> bool:
        """Return negated result from internal filter."""
        return not self.inner_filter.filter(model)


class TypeFilter(Filter):
    """Filter messages based on their type.

    Attributes:
        types: Container with accepted types.
    """

    def __init__(self, types: Container[Optional[str]]):
        self.types = types

    def filter(self, model: Record) -> bool:
        """Return True if message's type matches at least one of specified types."""
        return model.type in self.types


class AddressCountryFilter(Filter):
    """Filter letters based on recipient's country.

    Attributes:
        countries: Container with accepted countries.
    """

    def __init__(self, countries: Container[Optional[str]]):
        self.countries = countries

    def filter(self, model: Record) -> bool:
        """Return True if letter recipient's country matches at least one of specified countries."""
        return model.recipient.country in self.countries


class FileExistsFilter(Filter):
    """Filter messages based on a presence of a file.

    Attributes:
        path: Path to the lock file.
    """

    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def filter(self, model: Record) -> bool:
        """Return True if the lock file exist."""
        return self.path.is_file()
