"""Models for blocked addresses."""
from sqlalchemy import CheckConstraint, Column, Text, UniqueConstraint

from .base import BaseModel


class BlockedEmail(BaseModel):
    """Represents a blocked email address."""

    email = Column(Text, nullable=False, unique=True)

    def __str__(self) -> str:
        return self.email or ''


class BlockedPhone(BaseModel):
    """Represents a blocked phone number."""

    phone = Column(Text, nullable=False, unique=True)

    def __str__(self) -> str:
        return self.phone or ''


ADDRESS_PART_CHECK_SQL = ('name IS NOT NULL OR organization IS NOT NULL OR street IS NOT NULL OR city IS NOT NULL OR '
                          'state IS NOT NULL OR postal_code IS NOT NULL or country IS NOT NULL')


class BlockedAddress(BaseModel):
    """Represents a blocked address."""

    __table_args__ = (
        # At least one part of address needs to be defined.
        CheckConstraint(ADDRESS_PART_CHECK_SQL, name='part_required'),
        # Ensure addresses are unique.
        UniqueConstraint('name', 'organization', 'street', 'city', 'state', 'postal_code', 'country'),
    )

    name = Column(Text())
    organization = Column(Text())
    street = Column(Text(), doc="May contain multiple lines")
    city = Column(Text())
    state = Column(Text(), doc="State or province name")
    postal_code = Column(Text())
    country = Column(Text(), doc="Country name, not code")

    def __str__(self) -> str:
        values = (self.name, self.organization, self.street, self.city, self.state, self.postal_code, self.country)
        return ', '.join(v for v in values if v)
