"""Models for email messages."""
from email.utils import make_msgid
from uuid import UUID

from sqlalchemy import ARRAY, CheckConstraint, Column, ForeignKey, ForeignKeyConstraint, Integer, Text
from sqlalchemy.orm import relationship
from sqlalchemy_utils import JSONType, UUIDType

from ..settings import SETTINGS
from .base import BaseModel
from .message import ArchiveReferenceMixin, ReferenceMixin, TemplateArchiveMessageMixin, TemplateMessageMixin
from .utils import ScalarListType


def make_message_id() -> str:
    """Return a new Message-ID."""
    # Separate function to make lazy usage of settings.
    return make_msgid(domain=SETTINGS['msgid_domain'])


class EmailMixin:
    """Shared columns between email and archive email."""

    # Envelope attributes
    message_id = Column(Text(), nullable=False, unique=True, default=make_message_id,
                        doc="RFC 2822-compliant Message-ID header.")

    # Message attributes
    sender = Column(Text())
    recipient = Column(ARRAY(Text).with_variant(ScalarListType(), 'sqlite'), nullable=False)
    extra_headers = Column(JSONType(), doc="Additional email message headers.")
    subject_template = Column(Text(), nullable=False, doc="Name of the subject template.")
    subject_template_uuid = Column(UUIDType(binary=False))
    body_template_html = Column(Text(), doc="Name of the HTML body template.")
    body_template_html_uuid = Column(
        UUIDType(binary=False),
        CheckConstraint('body_template_html_uuid IS NULL OR body_template_html IS NOT NULL',
                        name='body_template_html_uuid'))
    attachments = Column(ARRAY(UUIDType).with_variant(ScalarListType(UUID), 'sqlite'),
                         doc="List of attachment UUIDs.")


class Email(EmailMixin, TemplateMessageMixin, BaseModel):
    """Represents an email message."""

    references = relationship('EmailReference')

    def __repr__(self) -> str:
        recipients = []
        for recipient in self.recipient:
            if not len(recipient) > 61:
                recipients.append(recipient)
                continue
            prefix, suffix = recipient.split('@', 1)
            before_sign = prefix[:30] + '...' if len(prefix) > 30 else prefix
            after_sign = '...' + suffix[-30:] if len(suffix) > 30 else suffix
            recipients.append(before_sign + '@' + after_sign)
        return '<{} {} to {}>'.format(type(self).__name__, self.uuid, recipients)


class EmailReference(ReferenceMixin, BaseModel):
    """Represents an email message reference."""

    message_id = Column(ForeignKey(Email.__table__.columns.id), nullable=False)  # type: ignore[attr-defined]


class ArchiveEmail(EmailMixin, TemplateArchiveMessageMixin, BaseModel):
    """Represents an archived email message."""

    # Message-ID may be NULL in old emails.
    message_id = Column(Text(), nullable=True, default=make_message_id, index=True,
                        doc="RFC 2822-compliant Message-ID header.")
    references = relationship('ArchiveEmailReference')


class ArchiveEmailReference(ArchiveReferenceMixin, BaseModel):
    """Represents an archive email message reference."""

    __table_args__ = (
        ForeignKeyConstraint(
            ['message_id', 'create_datetime'],
            [ArchiveEmail.__table__.columns.id,  # type: ignore[attr-defined]
             ArchiveEmail.__table__.columns.create_datetime]),  # type: ignore[attr-defined]
    ) + ArchiveReferenceMixin.__table_args__

    message_id = Column(Integer, nullable=False)
