"""Models for letter messages."""
from dataclasses import astuple, dataclass
from typing import Optional

from sqlalchemy import Column, ForeignKey, ForeignKeyConstraint, Integer, Text
from sqlalchemy.orm import relationship
from sqlalchemy_utils import UUIDType

from .base import BaseModel
from .message import ArchiveMessageMixin, ArchiveReferenceMixin, MessageMixin, ReferenceMixin


@dataclass(frozen=True)
class Address:
    """Represents letter address."""

    name: Optional[str] = None
    organization: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    postal_code: Optional[str] = None
    country: Optional[str] = None

    def __str__(self) -> str:
        chunks = (i for i in astuple(self) if i is not None)
        return ', '.join(chunks)


class LetterMixin:
    """Shared columns between letter and archive letter."""

    # Message attributes
    recipient_name = Column(Text())
    recipient_organization = Column(Text())
    recipient_street = Column(Text(), doc="May contain multiple lines")
    recipient_city = Column(Text())
    recipient_state = Column(Text(), doc="State or province name")
    recipient_postal_code = Column(Text())
    recipient_country = Column(Text(), doc="Country name, not code")
    file = Column(UUIDType(binary=False), nullable=False)
    # Send attributes
    batch_id = Column(Text(), doc="An identifier of a batch that sent the letter.")

    @property
    def recipient(self) -> Address:
        """Return address object."""
        return Address(name=self.recipient_name, organization=self.recipient_organization, street=self.recipient_street,
                       city=self.recipient_city, state=self.recipient_state, postal_code=self.recipient_postal_code,
                       country=self.recipient_country)

    @recipient.setter
    def recipient(self, value: Address) -> None:
        self.recipient_name = value.name
        self.recipient_organization = value.organization
        self.recipient_street = value.street
        self.recipient_city = value.city
        self.recipient_state = value.state
        self.recipient_postal_code = value.postal_code
        self.recipient_country = value.country


class Letter(LetterMixin, MessageMixin, BaseModel):  # type: ignore[misc, unused-ignore] # Ignore signature on recipient
    """Represents a letter."""

    references = relationship('LetterReference')


class LetterReference(ReferenceMixin, BaseModel):
    """Represents a letter reference."""

    message_id = Column(ForeignKey(Letter.__table__.columns.id), nullable=False)  # type: ignore[attr-defined]


class ArchiveLetter(LetterMixin, ArchiveMessageMixin, BaseModel):  # type: ignore[misc, unused-ignore]
    """Represents an archived letter."""

    references = relationship('ArchiveLetterReference')


class ArchiveLetterReference(ArchiveReferenceMixin, BaseModel):
    """Represents an archive letter message reference."""

    __table_args__ = (
        ForeignKeyConstraint(
            ['message_id', 'create_datetime'],
            [ArchiveLetter.__table__.columns.id,  # type: ignore[attr-defined]
             ArchiveLetter.__table__.columns.create_datetime]),  # type: ignore[attr-defined]
    ) + ArchiveReferenceMixin.__table_args__

    message_id = Column(Integer, nullable=False)
