"""Messenger settings management."""
import logging
import os
from logging.config import dictConfig
from typing import Any, Iterable, Optional, Sequence, Tuple, cast

import yaml
from pydantic import BaseSettings, PrivateAttr

try:  # pragma: no cover
    import sentry_sdk
    from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
except ImportError:
    sentry_sdk = None
    SqlalchemyIntegration = None

_LOGGER = logging.getLogger(__name__)
ENVIRON_NAME = 'MESSENGER_CONFIG'
CONFIG_FILES = ('~/.fred/messenger.conf', '/etc/fred/messenger.conf')


class MessengerSettings(BaseSettings):
    """Actual settings."""

    clean_days: int = 30
    db_connection: str = 'sqlite:///:memory:'
    db_echo: bool = False
    delivery_backend: str = 'messenger.delivery.email.ImapDelivery'
    delivery_options: dict = {}
    processors: dict = {}
    fileman_netloc: Optional[str] = None
    fred_db_connection: Optional[str] = None
    logging: Optional[dict] = None
    grpc_max_workers: Optional[int] = None
    grpc_maximum_concurrent_rpcs: Optional[int] = None
    grpc_options: Optional[Sequence] = None
    grpc_port: int = 50051
    monitoring_email: Optional[str] = None
    msgid_domain: Optional[str] = None
    senders: dict = {}
    sentry: dict = {}
    test_db_connection: str = 'sqlite:///:memory:'
    # Pseudo-setting which holds a command verbosity level.
    _verbosity: int = PrivateAttr(1)


class Settings:
    """Settings manager - wraps messenger settings to take care of loading.."""

    def __init__(self) -> None:
        self._settings: Optional[MessengerSettings] = None

    def load(self, config_argv: Optional[str] = None) -> None:
        """Return setting object from the first config file found.

        Raises:
            FileNotFoundError: If the config file is provided, but the file doesn't exist.
            RuntimeError: If settings are already loaded.
        """
        if self._settings is not None:
            raise RuntimeError('Settings already loaded')
        config_file = os.environ.get(ENVIRON_NAME, config_argv)
        if not config_file:
            for name in CONFIG_FILES:
                full_name = os.path.expanduser(name)
                if os.path.isfile(full_name):
                    config_file = full_name
                    break
        if config_file:
            with open(config_file) as file:
                settings = yaml.safe_load(file)
        else:
            settings = {}
        self._settings = MessengerSettings(**settings)
        self._setup_logs()

    def _setup_logs(self) -> None:
        """Set up logging systems."""
        # Apply logging configuration
        if self['logging'] is not None:
            dictConfig(self['logging'])
        # Set up Sentry
        if self['sentry'] and self['sentry'].get('dsn'):
            if sentry_sdk is None:
                _LOGGER.warning('Sentry set up, but sentry-sdk not available.')
            else:
                sentry_sdk.init(dsn=self['sentry']['dsn'], integrations=[SqlalchemyIntegration()],
                                environment=self['sentry'].get('environment'), ca_certs=self['sentry'].get('ca_certs'))

    def __getitem__(self, key: str) -> Any:
        if self._settings is None:
            self.load()
        try:
            return getattr(self._settings, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        if self._settings is None:
            self.load()
        try:
            setattr(self._settings, key, value)
        except ValueError:
            raise KeyError(key)

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Return an iterator of pairs of all settings."""
        if self._settings is None:
            self.load()
        return cast(MessengerSettings, self._settings).dict().items()


class SettingsProxy:
    """Proxy for settings instance.

    Attributes:
        settings: Actual settings
    """

    def __init__(self, settings: Settings):
        self.settings = settings

    def __getattr__(self, name: str) -> Any:
        return getattr(self.settings, name)

    def __getitem__(self, key: str) -> Any:
        return self.settings[key]

    def __setitem__(self, key: str, value: Any) -> Any:
        self.settings[key] = value


SETTINGS = SettingsProxy(Settings())
