from datetime import datetime
from email.message import EmailMessage
from smtplib import SMTPNotSupportedError
from typing import Optional, Tuple, cast
from unittest import TestCase
from unittest.mock import Mock, call, patch
from uuid import UUID

from filed import File, FileClient, Storage
from freezegun import freeze_time
from pytz import UTC
from testfixtures import Comparison

from messenger.backends.letter import BaseLetterBackend, EmailLetterBackend, LetterMessage
from messenger.constants import MessageStatus
from messenger.models import Address, Letter


class TestLetterBackend(BaseLetterBackend):
    """Simple backend for tests of BaseLetterBackend."""

    def _send(self, message: LetterMessage) -> None:
        pass  # pragma: no cover


class BaseLetterBackendTest(TestCase):
    uuid = UUID(int=42)
    recipient = Address(name='Kryten 2X4B-523P')
    file_uuid = UUID(int=255)

    def setUp(self):
        self.file_client = Mock(FileClient)
        self.backend = TestLetterBackend(storage=Storage(self.file_client))
        self.file = File(uuid=self.file_uuid, client=self.file_client)

    def get_letter(self, file_uuid: UUID = file_uuid) -> Letter:
        return Letter(uuid=self.uuid, recipient=self.recipient, file=file_uuid)

    def _test_get_message(self, letter: Letter, sender: Optional[Address] = None) -> LetterMessage:
        message = self.backend.get_message(letter)

        self.assertIsInstance(message, LetterMessage)
        self.assertEqual(message.uuid, self.uuid)
        self.assertEqual(message.recipient, self.recipient)
        self.assertEqual(message.file.uuid, self.file_uuid)
        self.assertEqual(message.sender, sender)
        return message

    def test_init_sender(self):
        backend = TestLetterBackend(sender={'name': 'Holly'})
        self.assertEqual(backend.sender, Address(name='Holly'))

    def test_init_invalid_address(self):
        with self.assertRaisesRegex(ValueError, "Invalid sender: .*"):
            TestLetterBackend(sender={'invalid': 'Holly'})

    def test_get_message(self):
        letter = self.get_letter()
        self._test_get_message(letter)

    def test_get_message_sender(self):
        sender = {'name': 'Holly'}
        sender_address = Address(name='Holly')
        self.backend = TestLetterBackend(sender=sender, storage=Storage(self.file_client))
        letter = self.get_letter()

        self._test_get_message(letter, sender=sender_address)


class EmailLetterBackendTest(TestCase):
    uuid = UUID(int=42)
    recipient = Address(name='Kryten 2X4B-523P')
    file_uuid = UUID(int=255)
    file_name = 'example.txt'
    file_mimetype = 'text/plain'
    file_content = 'Gazpacho!'
    to_address = 'kryten@example.org'

    def setUp(self):
        self.file_client = Mock(FileClient)
        self.backend = TestLetterBackend(storage=Storage(self.file_client))
        self.file = File(uuid=self.file_uuid, client=self.file_client)
        self.file_client.stat.return_value = {'mimetype': self.file_mimetype, 'name': self.file_name}
        self.file_client.read.return_value = [self.file_content.encode('utf-8')]

    def get_letter(self, file_uuid: UUID = file_uuid) -> Letter:
        return Letter(uuid=self.uuid, recipient=self.recipient, file=file_uuid, attempts=0,
                      status=MessageStatus.PENDING)

    def _test_send(self, backend: EmailLetterBackend, *, from_address: str = 'webmaster@localhost',
                   subject: str = 'A letter from messenger',
                   body: str = 'A letter to Kryten 2X4B-523P, see attachment.\n') -> None:
        letter = self.get_letter()

        with patch('smtplib.SMTP', autospec=True) as smtp_mock:
            backend.send(letter)

        calls = [call('localhost', 25), call().send_message(Comparison(EmailMessage))]
        self.assertEqual(smtp_mock.mock_calls, calls)
        email = smtp_mock.mock_calls[1][1][0]
        self.assertEqual(email['To'], self.to_address)
        self.assertEqual(email['From'], from_address)
        self.assertEqual(email['Subject'], subject)
        self.assertTrue(email.is_multipart())
        self.assertEqual(email.get_body('plain').get_content(), body)

        attachments = cast(Tuple[EmailMessage], tuple(email.iter_attachments()))
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0].get_content_type(), self.file_mimetype)
        self.assertEqual(attachments[0].get_filename(), self.file_name)
        self.assertEqual(attachments[0].get_content(), self.file_content)

    def test_send(self):
        backend = EmailLetterBackend(to_address=self.to_address, storage=Storage(self.file_client))
        self._test_send(backend)

    def test_send_custom_from(self):
        from_address = 'holly@example.org'
        backend = EmailLetterBackend(to_address=self.to_address, from_address=from_address,
                                     storage=Storage(self.file_client))
        self._test_send(backend, from_address=from_address)

    def test_send_custom_subject(self):
        subject = 'Message: {letter.uuid}'
        backend = EmailLetterBackend(to_address=self.to_address, subject=subject, storage=Storage(self.file_client))
        self._test_send(backend, subject='Message: 00000000-0000-0000-0000-00000000002a')

    def test_send_custom_body(self):
        body = 'Letter: {letter.file.name}'
        backend = EmailLetterBackend(to_address=self.to_address, body=body, storage=Storage(self.file_client))
        self._test_send(backend, body='Letter: example.txt\n')

    @freeze_time('2020-02-02 14:00:00')
    def test_send_smtp_error(self):
        backend = EmailLetterBackend(to_address=self.to_address, storage=Storage(self.file_client))
        letter = self.get_letter()

        with patch('smtplib.SMTP', autospec=True) as smtp_mock:
            smtp_mock.return_value.send_message.side_effect = SMTPNotSupportedError
            letter = backend.send(letter)

        calls = [call('localhost', 25), call().send_message(Comparison(EmailMessage))]
        self.assertEqual(smtp_mock.mock_calls, calls)
        # Check letter data
        self.assertEqual(letter.update_datetime, datetime(2020, 2, 2, 14, tzinfo=UTC))
        self.assertIsNone(letter.send_datetime)
        self.assertEqual(letter.attempts, 1)
        self.assertEqual(letter.status, MessageStatus.PENDING)
