from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterable, Tuple
from unittest import TestCase
from unittest.mock import sentinel

from messenger.filters import AddressCountryFilter, FileExistsFilter, NegativeFilter, TypeFilter
from messenger.models import Email, Letter, Record, Sms

from .utils import StaticFilter


class NegativeFilterTest(TestCase):
    def test_init_filter(self):
        inner = StaticFilter(True)
        filter = NegativeFilter(inner)
        self.assertIs(filter.inner_filter, inner)

    def test_init_str(self):
        filter = NegativeFilter('messenger.tests.utils.StaticFilter')
        self.assertIsInstance(filter.inner_filter, StaticFilter)

    def test_init_args(self):
        filter = NegativeFilter('messenger.tests.utils.StaticFilter', result=sentinel.result)
        self.assertIsInstance(filter.inner_filter, StaticFilter)
        self.assertEqual(filter.inner_filter.result, sentinel.result)

    def test_filter(self):
        data: Iterable[Tuple[NegativeFilter, Record, bool]] = (
            # filter, model, result
            # False to True
            (NegativeFilter(StaticFilter(True)), Email(), False),
            (NegativeFilter(StaticFilter(True)), Sms(), False),
            (NegativeFilter(StaticFilter(True)), Letter(), False),
            # True to False
            (NegativeFilter(StaticFilter(False)), Email(), True),
            (NegativeFilter(StaticFilter(False)), Sms(), True),
            (NegativeFilter(StaticFilter(False)), Letter(), True),
        )
        for filter, model, result in data:
            with self.subTest(filter=filter, model=model):
                self.assertEqual(filter.filter(model), result)


class TypeFilterTest(TestCase):
    def test_filter(self):
        data: Iterable[Tuple[TypeFilter, Record, bool]] = (
            # filter, model, result
            # Catch nothing
            (TypeFilter([]), Email(), False),
            (TypeFilter([]), Sms(), False),
            (TypeFilter([]), Letter(), False),
            (TypeFilter([]), Email(type=sentinel.type), False),
            (TypeFilter([]), Sms(type=sentinel.type), False),
            (TypeFilter([]), Letter(type=sentinel.type), False),
            # Match type
            (TypeFilter([sentinel.type]), Email(type=sentinel.type), True),
            (TypeFilter([sentinel.type]), Sms(type=sentinel.type), True),
            (TypeFilter([sentinel.type]), Letter(type=sentinel.type), True),
            (TypeFilter([sentinel.type]), Email(), False),
            (TypeFilter([sentinel.type]), Sms(), False),
            (TypeFilter([sentinel.type]), Letter(), False),
            # Match one of types
            (TypeFilter([sentinel.type1, sentinel.type2]), Email(type=sentinel.type1), True),
            (TypeFilter([sentinel.type1, sentinel.type2]), Sms(type=sentinel.type1), True),
            (TypeFilter([sentinel.type1, sentinel.type2]), Letter(type=sentinel.type1), True),
            (TypeFilter([sentinel.type1, sentinel.type2]), Email(type=sentinel.type2), True),
            (TypeFilter([sentinel.type1, sentinel.type2]), Sms(type=sentinel.type2), True),
            (TypeFilter([sentinel.type1, sentinel.type2]), Letter(type=sentinel.type2), True),
            (TypeFilter([sentinel.type1, sentinel.type2]), Email(), False),
            (TypeFilter([sentinel.type1, sentinel.type2]), Sms(), False),
            (TypeFilter([sentinel.type1, sentinel.type2]), Letter(), False),
            # Match None as type
            (TypeFilter([None]), Email(), True),
            (TypeFilter([None]), Sms(), True),
            (TypeFilter([None]), Letter(), True),
        )
        for filter, model, result in data:
            with self.subTest(filter=filter, model=model):
                self.assertEqual(filter.filter(model), result)


class AddressCountryFilterTest(TestCase):
    def test_filter(self):
        data = (
            # filter, model, result
            # Catch nothing
            (AddressCountryFilter([]), Letter(), False),
            (AddressCountryFilter([]), Letter(recipient_country=sentinel.country), False),
            # Match country
            (AddressCountryFilter([sentinel.country]), Letter(recipient_country=sentinel.country), True),
            (AddressCountryFilter([sentinel.country]), Letter(), False),
            # Match one of countries
            (AddressCountryFilter([sentinel.country1, sentinel.country2]),
             Letter(recipient_country=sentinel.country1), True),
            (AddressCountryFilter([sentinel.country1, sentinel.country2]),
             Letter(recipient_country=sentinel.country2), True),
            (AddressCountryFilter([sentinel.country1, sentinel.country2]), Letter(), False),
            # Match None as country
            (AddressCountryFilter([None]), Letter(), True),
        )
        for filter, model, result in data:
            with self.subTest(filter=filter, model=model):
                self.assertEqual(filter.filter(model), result)


class FileExistsFilterTest(TestCase):
    def setUp(self):
        self.tmp_dir = TemporaryDirectory()
        self.path = Path(self.tmp_dir.name) / 'lock'

    def tearDown(self):
        self.tmp_dir.cleanup()

    def test_file_present(self):
        self.path.touch()
        for path in (self.path, str(self.path)):
            with self.subTest(path=path):
                filter = FileExistsFilter(path)
                self.assertTrue(filter.filter(sentinel.model))

    def test_file_absent(self):
        filter = FileExistsFilter(self.path)
        self.assertFalse(filter.filter(sentinel.model))
