import os
import tempfile
from contextlib import contextmanager
from typing import Any, Dict
from unittest import TestCase
from unittest.mock import call, mock_open, patch

import yaml
from testfixtures import LogCapture

from messenger.settings import MessengerSettings, Settings, SettingsProxy

from .utils import TestSettings


@contextmanager
def set_environ(**kwargs):
    """Clean the environment and set only provided variables."""
    with patch.dict(os.environ) as mock:
        os.environ.clear()
        os.environ.update(kwargs)
        yield mock


class SettingsTest(TestCase):
    home_call = call('/home/holly/.fred/messenger.conf')
    system_call = call('/etc/fred/messenger.conf')

    def setUp(self):
        self.log_handler = LogCapture('messenger', propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()

    def test_getitem(self):
        settings = Settings()
        settings._settings = MessengerSettings(db_connection='Gazpacho!')
        self.assertEqual(settings['db_connection'], 'Gazpacho!')

    def test_getitem_unknown(self):
        settings = Settings()
        settings._settings = MessengerSettings()
        with self.assertRaisesRegex(KeyError, 'unknown'):
            settings['unknown']

    def test_setitem(self):
        settings = Settings()
        settings._settings = MessengerSettings(db_connection='Gazpacho!')
        settings['db_connection'] = 'Wine cellar'
        self.assertEqual(settings['db_connection'], 'Wine cellar')

    def test_setitem_unknown(self):
        settings = Settings()
        settings._settings = MessengerSettings()
        with self.assertRaisesRegex(KeyError, 'unknown'):
            settings['unknown'] = 'Wine cellar'

    def test_items_defaults(self):
        settings = Settings()
        settings._settings = MessengerSettings()
        self.assertEqual(settings.items(), MessengerSettings().dict().items())

    def test_items(self):
        settings = Settings()
        settings._settings = MessengerSettings(db_connection='Gazpacho!')
        self.assertEqual(settings.items(), MessengerSettings(db_connection='Gazpacho!').dict().items())

    def test_items_load(self):
        # Test `items` trigger `load`.
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'db_connection': 'Gazpacho!'}, config_io)

        settings = Settings()
        with set_environ(MESSENGER_CONFIG=config_file.name):
            items = settings.items()

        self.assertIn(('db_connection', 'Gazpacho!'), items)

    def assertDefaults(self, settings: Settings) -> None:
        """Check settings contain all default settings."""
        for key, value in MessengerSettings().dict().items():
            self.assertEqual(settings[key], value)

    def test_load_environ(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'db_connection': 'Gazpacho!'}, config_io)

        settings = Settings()
        with set_environ(MESSENGER_CONFIG=config_file.name):
            settings.load()

        self.assertEqual(settings['db_connection'], 'Gazpacho!')

    def test_load_environ_unknown(self):
        settings = Settings()
        with set_environ(MESSENGER_CONFIG='/does/not/exist'):
            with self.assertRaisesRegex(FileNotFoundError, '/does/not/exist'):
                settings.load()

    def test_load_argv(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'db_connection': 'Wine cellar'}, config_io)

        settings = Settings()
        with set_environ():
            settings.load(config_file.name)

        self.assertEqual(settings['db_connection'], 'Wine cellar')

    def test_load_argv_unknown(self):
        settings = Settings()
        with set_environ():
            with self.assertRaisesRegex(FileNotFoundError, '/does/not/exist'):
                settings.load('/does/not/exist')

    def test_load_home(self):
        data = yaml.dump({'db_connection': 'Rusty gate'})

        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('os.path.isfile', return_value=True):
                with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                    settings.load()

        self.assertEqual(settings['db_connection'], 'Rusty gate')
        self.assertIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)

    def test_load_system(self):
        data = yaml.dump({'db_connection': 'Androids'})

        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('os.path.isfile', side_effect=[False, True]):
                with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                    settings.load()

        self.assertEqual(settings['db_connection'], 'Androids')
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertIn(self.system_call, open_mock.mock_calls)

    def test_load_no_settings(self):
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('os.path.isfile', return_value=False):
                with patch('builtins.open', side_effect=FileNotFoundError) as open_mock:
                    settings.load()

        self.assertDefaults(settings)
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)
        # Check no logs were made.
        self.log_handler.check()

    def test_load_error(self):
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('builtins.open', side_effect=OSError) as open_mock:
                settings.load()

        self.assertDefaults(settings)
        self.assertNotIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)

    def test_load_no_reload(self):
        # Test reload is not allowed.
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'db_connection': 'Gazpacho!'}, config_io)

        settings = Settings()
        with set_environ(MESSENGER_CONFIG=config_file.name):
            settings.load()
            with self.assertRaisesRegex(RuntimeError, 'Settings already loaded'):
                settings.load()

    def _test_load(self, config: Dict[str, Any]) -> Settings:
        # Generic utility to load settings for tests of `load` method.
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump(config, config_io)

        settings = Settings()
        with set_environ(MESSENGER_CONFIG=config_file.name):
            settings.load()

        return settings

    def test_load_logging(self):
        config = {'logging_key': 'logging_value'}
        with patch('messenger.settings.dictConfig') as config_mock:
            self._test_load({'logging': config})

        self.assertEqual(config_mock.mock_calls, [call(config)])

    def test_load_no_sentry(self):
        with patch('messenger.settings.sentry_sdk') as sentry_mock:
            with patch('messenger.settings.SqlalchemyIntegration') as si_mock:
                self._test_load({})

        self.assertEqual(sentry_mock.mock_calls, [])
        self.assertEqual(si_mock.mock_calls, [])

    def test_load_sentry(self):
        dsn = 'SMEGHEAD!'
        with patch('messenger.settings.sentry_sdk') as sentry_mock:
            with patch('messenger.settings.SqlalchemyIntegration') as si_mock:
                self._test_load({'sentry': {'dsn': dsn}})

        self.assertEqual(
            sentry_mock.mock_calls,
            [call.init(dsn=dsn, integrations=[si_mock.return_value], environment=None, ca_certs=None)])

    def test_load_sentry_missing(self):
        dsn = 'SMEGHEAD!'
        with patch('messenger.settings.sentry_sdk', new=None):
            self._test_load({'sentry': {'dsn': dsn}})

        self.log_handler.check(('messenger.settings', 'WARNING', 'Sentry set up, but sentry-sdk not available.'))

    def test_load_sentry_environment(self):
        dsn = 'SMEGHEAD!'
        environment = 'JMC'
        with patch('messenger.settings.sentry_sdk') as sentry_mock:
            with patch('messenger.settings.SqlalchemyIntegration') as si_mock:
                self._test_load({'sentry': {'dsn': dsn, 'environment': environment}})

        self.assertEqual(
            sentry_mock.mock_calls,
            [call.init(dsn=dsn, integrations=[si_mock.return_value], environment=environment, ca_certs=None)])

    def test_load_sentry_ca_certs(self):
        dsn = 'SMEGHEAD!'
        ca_certs = 'Silver Swimming Certificate'
        with patch('messenger.settings.sentry_sdk') as sentry_mock:
            with patch('messenger.settings.SqlalchemyIntegration') as si_mock:
                self._test_load({'sentry': {'dsn': dsn, 'ca_certs': ca_certs}})

        self.assertEqual(
            sentry_mock.mock_calls,
            [call.init(dsn=dsn, integrations=[si_mock.return_value], environment=None, ca_certs=ca_certs)])

    def test_getitem_load(self):
        # Test settings loads if not already loaded.
        data = yaml.dump({'db_connection': 'Gazpacho!'})
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('os.path.isfile', return_value=True):
                with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                    self.assertEqual(settings['db_connection'], 'Gazpacho!')

        self.assertIn(self.home_call, open_mock.mock_calls)

    def test_setitem_load(self):
        # Test settings loads if not already loaded.
        data = yaml.dump({'db_connection': 'Gazpacho!'})
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('os.path.isfile', return_value=True):
                with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                    settings['db_connection'] = 'Wine cellar'

        self.assertEqual(settings['db_connection'], 'Wine cellar')
        self.assertIn(self.home_call, open_mock.mock_calls)


class SettingsProxyTest(TestCase):
    def test_getitem(self):
        settings = TestSettings()
        settings['db_connection'] = 'Gazpacho!'
        proxy = SettingsProxy(settings)
        self.assertEqual(proxy['db_connection'], 'Gazpacho!')

    def test_setitem(self):
        settings = TestSettings()
        proxy = SettingsProxy(settings)
        proxy['db_connection'] = 'Gazpacho!'
        self.assertEqual(settings['db_connection'], 'Gazpacho!')

    def test_getattr(self):
        settings = TestSettings()
        settings['db_connection'] = 'Gazpacho!'
        proxy = SettingsProxy(settings)
        self.assertEqual(proxy.items(), settings.items())
